import numpy as np

"""ETS methods to use for initialization of parameters"""

def ANN(dep_var, init_components, params):

    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components
            
    alpha = params[0]

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            errors.append(0)

        else:               

            y_hat = lvl
            error = row.numpy()[0] - y_hat

            

            lvl = lvl + alpha*error

            errors.append(error)

    return errors

def ANA(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components      
    alpha, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl +seasonal
            error = row.numpy()[0] - y_hat

            lprev = lvl

            seasonal_t = seasonal+(gamma*error)
            lvl = lprev + alpha*error

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + seasonal
            error = row.numpy()[0] - y_hat

            lprev = lvl

            seasonal_t = seasonal+(gamma*error)
            lvl = lprev + alpha*error

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def ANM(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components          
    alpha, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl*seasonal
            error = row.numpy()[0] - y_hat

            lprev = lvl

            seasonal_t = seasonal+ gamma*error/lprev
            lvl = lprev + alpha*error/seasonal

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl*seasonal
            error = row.numpy()[0] - y_hat

            lprev = lvl

            seasonal_t = seasonal+ gamma*error/lprev
            lvl = lprev + alpha*error/seasonal

            seasonals.append(seasonal_t)
            errors.append(error)            

    return errors

def MNM(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl*seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev = lvl

            seasonal_t = seasonal*(1 + gamma*error)
            lvl = lprev*(1 + alpha*error)

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl*seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev = lvl

            seasonal_t = seasonal*(1 + gamma*error)
            lvl = lprev*(1 + alpha*error)

            seasonals.append(seasonal_t)
            errors.append(error)            

    return errors

def MNA(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev = lvl

            seasonal_t = seasonal + gamma*error*(lprev+seasonal)
            lvl = lprev + alpha*error*(lprev+ seasonal)

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev = lvl

            seasonal_t = seasonal + gamma*error*(lprev+seasonal)
            lvl = lprev + alpha*error*(lprev+ seasonal)

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors


def MNN(dep_var, init_components, params):

    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components            
    alpha = params[0]

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            errors.append(0)

        else:               

            y_hat = lvl
            error = (row.numpy()[0] - y_hat)/y_hat

            lvl = lvl*(1+alpha*error)

            errors.append(error)

    return errors

def AAdA(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + phi*trend + seasonal
            error = row.numpy()[0] - y_hat
            
            lprev, bprev = lvl, trend

            seasonal_t = seasonal + gamma*error
            lvl = lprev + phi*bprev + alpha*error
            trend = phi*bprev + beta*error


            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + phi*trend + seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal + gamma*error
            lvl = lprev + phi*bprev + alpha*error
            trend = phi*bprev + beta*error

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def MAdA(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + phi*trend + seasonal
            error = (row.numpy()[0] - y_hat)/y_hat
            
            lprev, bprev = lvl, trend

            seasonal_t = seasonal + gamma*error
            lvl = lprev + phi*bprev + alpha*error*(lprev + phi*bprev + seasonal)
            trend = phi*bprev + beta*error*(lprev + phi*bprev + seasonal)

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + phi*trend + seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal + gamma*error
            lvl = lprev + phi*bprev + alpha*error*(lprev + phi*bprev + seasonal)
            trend = phi*bprev + beta*error*(lprev + phi*bprev + seasonal)

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def AAdM(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = (lvl + phi*trend)*seasonal
            error = row.numpy()[0] - y_hat
            
            seasonal_t = seasonal + (gamma*error)/(lvl + phi*trend)
            lvl = lvl + phi*trend + alpha*error/seasonal
            trend = phi*trend + beta*error/seasonal

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = (lvl + phi*trend)*seasonal
            error = row.numpy()[0] - y_hat

            seasonal_t = seasonal + (gamma*error)/(lvl + phi*trend)
            lvl = lvl + phi*trend + alpha*error/seasonal
            trend = phi*trend + beta*error/seasonal

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def MAdM(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = (lvl + phi*trend)*seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal*(gamma*error + 1)
            lvl = (lprev + phi*bprev)*(1+alpha*error)
            trend = phi*bprev + beta*(lprev + phi*bprev)*error

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = (lvl + phi*trend)*seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal*(gamma*error + 1)
            lvl = (lprev + phi*bprev)*(1+alpha*error)
            trend = phi*bprev + beta*(lprev + phi*bprev)*error

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def AAA(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + trend +seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal+ gamma*error
            lvl = lprev + bprev+ alpha*error
            trend = bprev + beta*error

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + trend +seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal+ gamma*error
            lvl = lprev + bprev+ alpha*error
            trend = bprev + beta*error

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def MAA(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + trend + seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal+gamma*error*(lprev+bprev+seasonal)
            lvl = lprev+bprev+alpha*(lprev+bprev+seasonal)*error
            trend = bprev+ beta*(lprev+bprev+seasonal)*error

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + trend + seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal+gamma*error*(lprev+bprev+seasonal)
            lvl = lprev+bprev+alpha*(lprev+bprev+seasonal)*error
            trend = bprev+ beta*(lprev+bprev+seasonal)*error

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def AAN(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            errors.append(0) 

        else:               

            y_hat = lvl + trend 
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            lvl = lprev + bprev + alpha*error
            trend = bprev+ beta*error

            errors.append(error)

    return errors

def AAdN(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components            
    alpha, beta, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            errors.append(0) 

        else:               

            y_hat = lvl + phi*trend 
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            lvl = lprev + phi*bprev + alpha*error
            trend = phi*bprev+ beta*error

            errors.append(error)

    return errors


def MAN(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components            
    alpha, beta = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            errors.append(0) 

        else:               

            y_hat = lvl + trend 
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            lvl = (lprev+bprev)*(1+alpha*error)
            trend = bprev+ beta*(lprev+bprev)*error

            errors.append(error)

    return errors

def MAdN(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components            
    alpha, beta, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            errors.append(0) 

        else:               

            y_hat = lvl + phi*trend 
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            lvl = (lprev+phi*bprev)*(1+alpha*error)
            trend = phi*bprev+ beta*(lprev+phi*bprev)*error

            errors.append(error)

    return errors

def AAM(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = (lvl + trend)* seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal + gamma*error/(lprev+bprev)
            lvl = lprev+bprev + alpha*error/seasonal
            trend = bprev+ beta*error/seasonal

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               
            seasonal = seasonals[index-seasonal_periods]  

            y_hat = (lvl + trend)* seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal + gamma*error/(lprev+bprev)
            lvl = lprev+bprev + alpha*error/seasonal
            trend = bprev+ beta*error/seasonal

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors

def MAM(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = (lvl + trend)*seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal*(gamma*error + 1)
            lvl = (lprev + bprev)*(1+alpha*error)
            trend = bprev + beta*(lprev + bprev)*error

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = (lvl + trend)*seasonal
            error = (row.numpy()[0] - y_hat)/y_hat

            lprev, bprev = lvl, trend

            seasonal_t = seasonal*(gamma*error + 1)
            lvl = (lprev + bprev)*(1+alpha*error)
            trend = bprev + beta*(lprev + bprev)*error

            seasonals.append(seasonal_t)
            errors.append(error)

    return errors


import torch
from preprocessing.dataloader import DataLoader

"""Performance evaluation metrics for model predictions"""

def mae(y_pred, y_true):
    """
    Mean Absolute Error
    
    Arguments:
    
    *y_pred (array_like): Predicted values
    *y_true (array_like): Ground truth values
    """
    y_pred = DataLoader(y_pred).to_tensor()
    y_true = DataLoader(y_true).to_tensor()      

    return torch.mean(torch.abs(torch.sub(y_pred, y_true)))


def mse(y_pred, y_true):
    """
    Mean Squared Error
    
    Arguments:
    
    *y_pred (array_like): Predicted values
    *y_true (array_like): Ground truth values
    """
    y_pred = DataLoader(y_pred).to_tensor()
    y_true = DataLoader(y_true).to_tensor()      

    return torch.mean((torch.square(torch.sub(y_pred, y_true))))

def rmse(y_pred, y_true):
    """
    Root Mean Squared Error
    
    Arguments:
    
    *y_pred (array_like): Predicted values
    *y_true (array_like): Ground truth values
    """
    y_pred = DataLoader(y_pred).to_tensor()
    y_true = DataLoader(y_true).to_tensor()  

    return torch.sqrt(torch.mean((torch.square(torch.sub(y_pred, y_true)))))


import matplotlib
import matplotlib.style
import matplotlib.pyplot as plt
from preprocessing.dataloader import DataLoader

def plot_decomp(trend, seasonal, remainder, figsize=(12,8), colors=None, style=None):
    """
    Utility function for plotting time series decomposition results
    
    Arguments:

    *trend (array_like): Trend component of the decomposition
    *seasonal (array_like): Seasonal component of the decomposition
    *remainer (array_like): Remainders of the decomposition
    *figsize (Optional[tuple]): Size of the plot
    *colors (Optional[iterable]): Colors of the lines/points on the plot
    *style (Optional[str]): Style of the plot 'https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html'
    """
    if style:
        assert(type(style) == str), "Provide style as a string"
        matplotlib.style.use(style)
    
    use_colors = {"trend": "blue", "seasonal": "blue", "remainder": "blue"}

    if colors:
        try:
            iter(colors)
        except TypeError:
            raise TypeError("Provide colors as an iterable")
        
        if type(colors) == dict:
            for key in colors:
                assert(key in list(use_colors.keys())), f"Ensure that keys in colours dictionary are {list(use_colors.keys())}"
                use_colors[key] = colors[key]

        else:
            for ind, c in enumerate(colors):
                use_colors[list(use_colors.keys())[ind]] = c

    trend = DataLoader(trend).to_numpy()
    seasonal = DataLoader(seasonal).to_numpy()
    remainder = DataLoader(remainder).to_numpy()
    
    fig, axes = plt.subplots(3, 1, figsize=figsize)
    ax1, ax2, ax3 = axes     
    ax1.plot(range(len(trend)), trend, color=use_colors["trend"])
    ax1.set_ylabel("Trend")
    ax2.plot(range(len(seasonal)), seasonal, color=use_colors["seasonal"])
    ax2.set_ylabel("Seasonal")
    ax3.scatter(range(len(remainder)), remainder, color=use_colors["remainder"])
    ax3.set_ylabel("Remainder")
    plt.show()
        

def plot_train_test_split(train_data, test_data, val_data=None, figsize=(12,8), title: str =None, colors=None, style=None):
    """
    Utility function for plotting train test split
    
    Arguments:

    *train_data (array_like): Training data of the split
    *test_data (array_like): Test data of the split
    *val_data (Optional[array_like]): Val data of the split if data is splitted as train/val/test
    *figsize (Optional[tuple]): Size of the plot
    *title (Optional[str]): Title of the plot
    *colors (Optional[iterable]): Colors of the lines/points on the plot
    *style (Optional[str]): Style of the plot 'https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html'
    """
    
    assert (type(title) == "str" or title is None), "Plot title must be a string"

    if style:
        assert(type(style) == str), "Provide style as a string"
        matplotlib.style.use(style)
    
    use_colors = {"train": "blue", "val": "orange", "test": "red"}

    if colors:
        try:
            iter(colors)
        except TypeError:
            raise TypeError("Provide colors as an iterable")
        
        if type(colors) == dict:
            for key in colors:
                assert(key in list(use_colors.keys())), f"Ensure that keys in colours dictionary are {list(use_colors.keys())}"
                use_colors[key] = colors[key]

        else:
            for ind, c in enumerate(colors):
                use_colors[list(use_colors.keys())[ind]] = c

    train = DataLoader(train_data).to_numpy()
    test = DataLoader(test_data).to_numpy()

    if val_data:
        val = DataLoader(val_data).to_numpy()
        plt.figure(figsize=figsize)
        plt.plot(range(len(train)), train, label="Train", color=use_colors["train"])
        plt.plot(range(len(train), len(train+val_data)), val, label="Validation", color=use_colors["val"])
        plt.plot(range(len(train+val_data), len(train+val_data+test_data)), test, label="Test", color=use_colors["test"])
        plt.legend(loc="best")
        plt.title(title)
        plt.show()
       
    else:
        plt.figure(figsize=figsize)
        plt.plot(range(len(train)), train, label="Train", color=use_colors["train"])
        plt.plot(range(len(train), len(train+test)), test, label="Test", color=use_colors["test"])
        plt.legend(loc="best")
        plt.title(title)
        plt.show()
    


import numpy as np
import pandas as pd
import torch
from .dataloader import DataLoader
from scipy.stats import boxcox_normmax

class DataTransform:

    def __init__(self):
        """Base class for all data transformation class to inherit from"""
        pass

    def transform_assert(self, loader: DataLoader, axis: int, scales: dict):
        """Make necessary assertions for transforming the data"""
        for scale_key in scales:

            scale_dict = scales[scale_key]

            assert(type(scale_dict) == dict), f"Provide {scale_key} as a dict"

        assert (loader.to_tensor().numpy().size != 0), "Size of data must be > 0"

        if loader.data_type == pd.DataFrame:
            data_names = list(loader.organized_df.columns)
            for scale_dict in list(scales.values()):
                if scale_dict != {}:                
                    assert (len(list(scale_dict.keys())) == len(data_names)), "Provide same amount of entries as columns"
                    assert (set(list(scale_dict.keys())) == set(data_names)), "Provide with the same keys as column names"

        else:
            assert (loader.data.ndim <= 2), "Dimension of the data must be <= 2"
            assert (axis in [0, 1, -1, -2]), f"{axis} is not a valid axis"

            if loader.data.ndim == 1:
                assert (axis in [0,-1]), f"{axis} is not a valid axis of data of ndim == 1"
                
                for scale_key in scales:    
                    scale_dict = scales[scale_key]
                    if scale_dict != {}:
                        assert (list(scale_dict.keys()) == [scale_key]), "For data with ndim == 1, scales should be\
 provided as {}".format({f'{scale_key}': 'value'})

            else:
                hmap = {0: "row", 1: "col", -1: "col", -2: "row"}
                data_names = [f"{hmap[axis]}{i}" for i in range(loader.data.shape[axis])]

                for scale_dict in list(scales.values()):
                    if scales != {}:
                        assert (len(list(scale_dict.keys())) == len(data_names)), f"Provide same amount of entries as {hmap[axis]}s"
                        assert (set(list(scale_dict.keys())) == set(data_names)), f"Provide keys as {hmap[axis]}0 for {hmap[axis]} of index 0 etc..."

    def inv_transform_assert(self, loader: DataLoader, names, scales: dict):
        """Make necessary assertions to inverse transforming the data"""
        assert(type(names) == list), "Provide names argument as a list"

        try:
            if self.transformed_axis:
                pass
        except NameError:
            raise NameError("Must call '.transform()' before calling '.inverse_transform()'")

        assert(type(names) == list), "Provide names argument as a list"

        assert (loader.to_tensor().numpy().size != 0), "Size of data must be > 0"

        transform_data = loader.to_tensor()

        if names != []:
            for scale_dict in list(scales.values()):
                for name in names:
                    assert(name in list(scale_dict())), f"{name} was not in transformed data"
        else:
            assert (len(loader.to_tensor().shape) == 2), "Data must be 2 dimensional if the names argument is not provided"
            for scale_dict in list(scales.values()):
                assert (loader.to_tensor().shape[self.transformed_axis] == len(list(scales.keys()))), f"Expecting\
 size of axis {self.transformed_axis} = {len(list(scale_dict))} if the names argument is not provided"


class BoxCox(DataTransform):
    def __init__(self):
        """Box-Cox transformation for time series data"""
        super().__init__()
        self.lambdas = {}

    def transform(self, data, axis=-1, lambda_values: dict = {}):
        """
        Transform given data with box-cox method

        Arguments:

        *data (array_like): Time series data to transform
        *axis (int): Axis to perform the transformation on
        *lambda_values (Optional[dict]): Lambda values for the transformation; will be estimated if given as an empty dict 
        """

        loader = DataLoader(data)
        self.transform_assert(loader, axis, scales={"lambda": lambda_values})

        if loader.data_type == pd.DataFrame:
            data_names = list(loader.organized_df.columns)
            if lambda_values == {}:
                lambda_values = {name: None for name in data_names}
        else:
            if loader.data.ndim == 1:
                if lambda_values == {}:
                    lambda_values = {"lambda": None}
            else:
                hmap = {0: "row", 1: "col", -1: "col", -2: "row"}
                data_names = [f"{hmap[axis]}{i}" for i in range(loader.data.shape[axis])]
                if lambda_values == {}:
                    lambda_values = {name: None for name in data_names}

        self.transformed_axis = axis
        transform_data = loader.to_tensor()
        transformed = np.zeros(transform_data.shape)
        for ind in range(len(data_names)):

            if transform_data.ndim == 1:
                current_data = transform_data

            elif axis in [-1,1]:
                current_data = transform_data[:, ind]
            else:
                current_data = transform_data[ind, :]
            
            name = data_names[ind]
            lambd = lambda_values[name]
            boxcoxed = self.__box_cox(current_data, lambd, name)

            if transform_data.ndim == 1:
                transformed = boxcoxed

            elif axis in [-1,1]:
                transformed[:, ind] = boxcoxed
            
            else:
                transformed[ind, :] = boxcoxed

        return transformed
    
    def inverse_transform(self, data, names: list = []):
        """
        Inverse transform given data

        Arguments:

        *data (array_like): Time series data to inverse transform
        *names (Optional[list]): Keys for data to inverse transform if partial transformation is desired  
        """
        loader = DataLoader(data)

        self.inv_transform_assert(loader, names, scales={"lambda": self.lambdas})

        transform_data = loader.to_tensor()
        transformed = np.zeros(transform_data.shape)

        if names == []:
            names = list(self.lambdas.keys())

        for ind in range(len(names)):

            if transform_data.ndim == 1:
                current_data = transform_data

            elif self.transformed_axis in [-1,1]:
                current_data = transform_data[:, ind]
            
            else:
                current_data = transform_data[ind, :]
            
            name = names[ind]
            lambd = self.lambdas[name]

            inversed = torch.exp(current_data) if lambd == 0 else torch.pow(torch.add(torch.mul(lambd, current_data), 1), 1/lambd)

            if transform_data.ndim == 1:
                transformed = inversed
            
            elif self.transformed_axis in [-1,1]:
                transformed[:, ind] = inversed
            
            else:
                transformed[ind, :] = inversed

        return transformed
          
    def __box_cox(self, data, lambd=None, data_name="col0"):
        """Perform the box-cox transformation"""
        box_coxed = data.detach().clone()

        if lambd:
            assert(type(lambd) == float or type(lambd) == int), "Provide lambda value as a float"

        else:  
            lambd = boxcox_normmax(box_coxed.numpy())
        
        box_coxed = torch.log(box_coxed) if lambd == 0 else torch.div(torch.sub(torch.pow(box_coxed, lambd), 1), lambd)
        self.lambdas[data_name] = lambd

        return box_coxed

class StandardScaling(DataTransform):
    def __init__(self):
        """Standard Scaling for time series data"""
        self.locations = {}
        self.scales = {}

    def transform(self, data, axis=-1, locations: dict = {}, scales: dict = {}):
        """
        Standard scale the given data

        Arguments:

        *data (array_like): Time series data to transform
        *axis (int): Axis to perform the transformation on
        *locations (Optional[dict]): Location values to be used for scaling the data; will be taken as the mean if not given
        *scales (Optional[dict]): Scale values to be used for scaling the data; will be taken as the std if not given
        """
        loader = DataLoader(data)
        self.transform_assert(loader, axis, scales={"location": locations, "scale": scales})

        if loader.data_type == pd.DataFrame:
            data_names = list(loader.organized_df.columns)
            if locations == {}:
                locations = {name: loader.organized_df[name].mean() for name in data_names}
            if scales == {}:
                scales = {name: loader.organized_df[name].std() for name in data_names}
                for s in list(scales.values()):
                    if s == 0:
                        raise ValueError("cannot scale data with std = 0")
        else:
            if loader.data.ndim == 1:
                if locations == {}:
                    locations = {"loc": loader.to_tensor().mean().item()}
                if scales == {}:
                    scales = {"scale": loader.to_tensor().std().item()}
                    for s in list(scales.values()):
                        if s == 0:
                            raise ValueError("cannot scale data with std = 0")
            else:
                hmap = {0: "row", 1: "col", -1: "col", -2: "row"}
                data_names = [f"{hmap[axis]}{i}" for i in range(loader.data.shape[axis])]
                if locations == {}:
                    locs = loader.to_tensor().mean(axis)
                    locations = {data_names[ind]: locs[ind].item() for ind in range(len(data_names))}
                if scales == {}:
                    stds = loader.to_tensor().std(axis)
                    scales = {data_names[ind]: stds[ind].item() for ind in range(len(data_names))}
                    for s in list(scales.values()):
                        if s == 0:
                            raise ValueError("cannot scale data with std = 0")
        self.locations = locations
        self.scales = scales

        self.transformed_axis = axis
        transform_data = loader.to_tensor()
        transformed = np.zeros(transform_data.shape)
        for ind in range(len(data_names)):

            if transform_data.ndim == 1:
                current_data = transform_data

            elif axis in [-1,1]:
                current_data = transform_data[:, ind]
            
            else:
                current_data = transform_data[ind, :]
            
            name = data_names[ind]
            mu = self.locations[name]
            sigma = self.scales[name]
            
            x = current_data.detach().clone()
            standard_scaled = torch.div(torch.sub(x,mu), sigma)

            if transform_data.ndim == 1:
                transformed = standard_scaled

            elif axis in [-1,1]:
                transformed[:, ind] = standard_scaled
            
            else:
                transformed[ind, :] = standard_scaled

        return transformed
    
    def inverse_transform(self, data, names: list = []):
        """
        Inverse transform given data

        Arguments:

        *data (array_like): Time series data to transform
        *names (Optional[list]): Keys for data to inverse transform if partial transformation is desired  
        """
        loader = DataLoader(data)

        self.inv_transform_assert(loader, names, scales={"lambda": self.lambdas})

        transform_data = loader.to_tensor()
        transformed = np.zeros(transform_data.shape)

        for ind in range(len(names)):

            if transform_data.ndim == 1:
                current_data = transform_data

            elif self.transformed_axis in [-1,1]:
                current_data = transform_data[:, ind]
            
            else:
                current_data = transform_data[ind, :]
            
            name = names[ind]
            mu = self.locations[name]
            sigma = self.scales[name]

            x = current_data.detach().clone()
            inversed = torch.add(torch.mul(x, sigma), mu)

            if transform_data.ndim == 1:
                transformed = inversed
            
            elif self.transformed_axis in [-1,1]:
                transformed[:, ind] = inversed
            
            else:
                transformed[ind, :] = inversed

        return transformed

class MinMaxScaling(DataTransform):
    def __init__(self, feature_range=(0,1)):
        """
        MinMax Scaling for time series data

        Arguments:

        *feature_range (Optional[iterable]): Value bounds for the data to be scaled on
        """
        try:
            iter(feature_range)
        except TypeError:
            raise TypeError("Provide feature_range as an iterable")

        assert(len(feature_range) == 2), "Provide feature_range as an iterable of length 2"

        self.lb, self.ub = feature_range

        assert(self.ub > self.lb), "Provide feature_range as (a,b) where b > a"

        self.mins = {}
        self.maxes = {}

    def transform(self, data, axis=-1):
        """
        MinMax scale the given data

        Arguments:

        *data (array_like): Time series data to transform
        *axis (int): Axis to perform the transformation on
        """
        loader = DataLoader(data)
        self.transform_assert(loader, axis, scales={})

        if loader.data_type == pd.DataFrame:
            data_names = list(loader.organized_df.columns)
            mins = {name: loader.organized_df[name].min() for name in data_names}
            maxes = {name: loader.organized_df[name].max() for name in data_names}

            for key in mins:

                if mins[key] == maxes[key]:
                    raise ValueError("Cannot scale with min(data)=max(data)")

        else:
            if loader.data.ndim == 1:

                mins = {"min": loader.to_tensor().min().item()}
                maxes = {"max": loader.to_tensor().max().item()}

                if mins["min"] == maxes["max"]:
                    raise ValueError("Cannot scale with min(data)=max(data)")
                
            else:
                hmap = {0: "row", 1: "col", -1: "col", -2: "row"}
                data_names = [f"{hmap[axis]}{i}" for i in range(loader.data.shape[axis])]
                
                mins_ = loader.to_tensor().min(axis)
                mins = {data_names[ind]: mins_[ind].item() for ind in range(len(data_names))}
                maxes_ = loader.to_tensor().max(axis)
                maxes = {data_names[ind]: maxes_[ind].item() for ind in range(len(data_names))}

                for key in mins:

                    if mins[key] == maxes[key]:
                        raise ValueError("Cannot scale with min(data)=max(data)")
                    
        self.mins = mins
        self.maxes = maxes

        self.transformed_axis = axis
        transform_data = loader.to_tensor()
        transformed = np.zeros(transform_data.shape)
        for ind in range(len(data_names)):

            if transform_data.ndim == 1:
                current_data = transform_data

            elif axis in [-1,1]:
                current_data = transform_data[:, ind]
            
            else:
                current_data = transform_data[ind, :]
            
            name = data_names[ind]
            xmin = self.mins[name]
            xmax = self.maxes[name]
            
            x = current_data.detach().clone()
            minmax_scaled = torch.div(torch.sub(x,xmin), torch.sub(xmax-xmin))
            minmax_scaled = torch.add(torch.mul(minmax_scaled, (self.ub-self.lb)), self.lb)

            if transform_data.ndim == 1:
                transformed = minmax_scaled

            elif axis in [-1,1]:
                transformed[:, ind] = minmax_scaled
            
            else:
                transformed[ind, :] = minmax_scaled

        return transformed

    def inverse_transform(self, data, names: list = []):
        """
        Inverse transform given data

        Arguments:

        *data (array_like): Time series data to transform
        *names (Optional[list]): Keys for data to inverse transform if partial transformation is desired  
        """
        loader = DataLoader(data)

        self.inv_transform_assert(loader, names, scales={})

        transform_data = loader.to_tensor()
        transformed = np.zeros(transform_data.shape)

        for ind in range(len(names)):

            if transform_data.ndim == 1:
                current_data = transform_data

            elif self.transformed_axis in [-1,1]:
                current_data = transform_data[:, ind]
            
            else:
                current_data = transform_data[ind, :]
            
            name = names[ind]
            xmin = self.mins[name]
            xmax = self.maxes[name]

            x = current_data.detach().clone()
            inversed = torch.div(torch.add(x, self.lb), (self.ub-self.lb))
            inversed = torch.add(torch.mul(inversed, torch.sub(xmax-xmin)), xmin)

            if transform_data.ndim == 1:
                transformed = inversed
            
            elif self.transformed_axis in [-1,1]:
                transformed[:, ind] = inversed
            
            else:
                transformed[ind, :] = inversed


import numpy as np
import torch
from scipy.optimize import least_squares
import pandas as pd
from preprocessing.dataloader import DataLoader
from .initialization import get_init_method, get_smooth_method
import scipy.stats as stats

class Model():
    def __init__(self,dep_var, **kwargs):
        """
        Base model class for all model classes to inherit from.
        Child classes are expected to implement their own .fit() and .predict() methods
        """
        
        self.dep_var = DataLoader(dep_var).to_tensor()
        self.set_kwargs(kwargs)

    
    def set_allowed_kwargs(self, kwargs: list):
        """This function sets the allowed keyword arguments for the model."""
        self.allowed_kwargs = kwargs


    def __check_kwargs(self, kwargs: dict):
          """This function checks if the keyword arguments are valid."""
          for (k,v) in kwargs.items():
              if k not in self.allowed_kwargs:
                  raise ValueError("{key} is not a valid keyword for this model".format(key = k))

    def set_kwargs(self, kwargs: dict):
        """This function sets the keyword arguments for the model."""
        self.__check_kwargs(kwargs)
        for (k,v) in kwargs.items():
            self.__setattr__(k,v)

    def fit(self, dep_var, indep_var, *args, **kwargs):
        #This function will be overriden by the child class.
        raise NotImplementedError("This function is not implemented yet.")

    def predict(self, dep_var, indep_var, *args, **kwargs):
        # This function will be overriden by the child class.
        raise NotImplementedError("This function is not implemented yet.")
    
class Smoothing_Model(Model):

    def __init__(self, dep_var, trend=None,  seasonal=None,  seasonal_periods=None, damped=False, initialization_method="heuristic", **kwargs):
        """
        Base class for exponential smoothing methods.
        All smoothing methods inherit from this class and is used for parameter initialization.
        
        Arguments:

        *dep_var (array_like): Univariate time series data
        *trend (Optional[str]): Trend component; None or "add"
        *seasonal (Optional[str]): Seasonal component; None, "add" or "mul"
        *seasonal_periods (Optional[int]): Cyclic period of the seasonal component; int or None if seasonal is None
        *damped (bool): Damp factor of the trend component; False if trend is None
        *initialization_method (str): Initialization method to use for the model parameters; "heuristic" or "mle"

        Keyword Arguments:

        ** alpha (float): Smoothing parameter for level component; takes values in (0,1)
        ** beta (float): Smoothing parameter for trend component; takes values in (0,1)
        ** phi (float): Damp factor for trend component; takes values in (0,1]
        ** gamma (float): Smoothing parameter for seasonal component; takes values in (0,1)
        
        All of the smoothing methods have been written as the below textbook as a reference:
        'Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
        and practice. OTexts, 2014.'
        """
        super().set_allowed_kwargs(["alpha", "beta", "phi", "gamma"])
        super().__init__(dep_var, **kwargs)

        self.trend = trend
        self.damped = damped
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods
        self.init_method = initialization_method

        self.method = get_smooth_method(error=None, trend=trend, damped=damped, seasonal=seasonal)
        
        self.params = {"alpha": 0.1, "beta": 0.01, "gamma": 0.01, "phi": 0.99}
        self.init_components = {"level": None, "trend": None, "seasonal": None}
        

    def __estimate_params(self, init_components, params):
        """Estimate the best parameters to use during fitting and forecasting for the smoothing model"""
        def func(x):
            if self.dep_var.ndim == 1:
                dep_var = torch.unsqueeze(self.dep_var, axis=-1)
            else:
                dep_var = self.dep_var
            errs = self.method(dep_var, init_components=init_components, params=x)

            return np.mean(np.square(np.array(errs)))

        estimated_params = least_squares(fun=func, x0 = params, bounds=(0,1)).x

        return estimated_params
    
    def initialize_params(self, initialize_params):
        """Initialize the components and the parameters to use during fitting and forecasting for the smoothing model"""
        self.initial_level, self.initial_trend, self.initial_seasonals = get_init_method(method=self.init_method)(self.dep_var,trend=self.trend, 
                                                                                                                  seasonal=self.seasonal, 
                                                                                                                  seasonal_periods=self.seasonal_periods)

        self.init_components["level"] = self.initial_level,
        self.init_components["trend"] = self.initial_trend,
        self.init_components["seasonal"] = np.expand_dims(self.initial_seasonals, axis=-1)

        init_params = {param: self.params[param] for param in initialize_params}

        params = self.__estimate_params(
                        init_components = [self.initial_level, self.initial_trend, self.initial_seasonals, self.seasonal_periods],
                        params = list(init_params.values()))
        
        for index, param in enumerate(params):

            init_params[list(init_params.keys())[index]] = param

        for param in self.params:

            if param in init_params:
                self.params[param] = init_params[param]

        if not self.damped:
            self.params["phi"] = 1


class ETS_Model(Model):

    def __init__(self, dep_var, error_type="add", trend=None,  seasonal=None, seasonal_periods=None, damped=False, initialization_method="heuristic", **kwargs):
        """
        Base class for ETS models
        All smoothing methods inherit from this class and is used for parameter initialization.
        
        Arguments:

        *dep_var (array_like): Univariate time series data
        *error_type (str): Type of error of the ETS model; "add" or "mul"
        *trend (Optional[str]): Trend component; None or "add"
        *seasonal (Optional[str]): Seasonal component; None, "add" or "mul"
        *seasonal_periods (Optional[int]): Cyclic period of the seasonal component; int or None if seasonal is None
        *damped (bool): Damp factor of the trend component; False if trend is None
        *initialization_method (str): Initialization method to use for the model parameters; "heuristic" or "mle"

        Keyword Arguments:

        ** alpha (float): Smoothing parameter for level component; takes values in (0,1)
        ** beta (float): Smoothing parameter for trend component; takes values in (0,1)
        ** phi (float): Damp factor for trend component; takes values in (0,1]
        ** gamma (float): Smoothing parameter for seasonal component; takes values in (0,1)
        
        All of the smoothing methods have been written as the below textbook as a reference:
        'Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
        and practice. OTexts, 2014.'
        """
        super().set_allowed_kwargs(["alpha", "beta", "phi", "gamma"])
        super().__init__(dep_var,  **kwargs)

        self.trend = trend
        self.damped = damped
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods
        self.error_type = error_type
        self.init_method = initialization_method

        self.method = get_smooth_method(error=error_type, trend=trend, damped=damped, seasonal=seasonal)

        self.params = {"alpha": 0.1, "beta": 0.01, "gamma": 0.01, "phi": 0.99}
        self.init_components = {"level": None, "trend": None, "seasonal": None}
        

    def __estimate_params(self, init_components, params):
        """Estimate the best parameters to use during fitting and forecasting for the smoothing model"""
        def func(x):
            if self.dep_var.ndim == 1:
                dep_var = torch.unsqueeze(self.dep_var, axis=-1)
            else:
                dep_var = self.dep_var
            errs = self.method(dep_var, init_components=init_components, params=x)

            return np.mean(np.square(np.array(errs)))

        estimated_params = least_squares(fun=func, x0 = params, bounds=(0,1)).x

        return estimated_params
    
    def initialize_params(self, initialize_params):
        """Initialize the components and the parameters to use during fitting and forecasting for the smoothing model"""
        self.initial_level, self.initial_trend, self.initial_seasonals = get_init_method(method=self.init_method)(self.dep_var,trend=self.trend, 
                                                                                                                  seasonal=self.seasonal, 
                                                                                                                  seasonal_periods=self.seasonal_periods)
        self.init_components["level"] = self.initial_level,
        self.init_components["trend"] = self.initial_trend,
        self.init_components["seasonal"] = np.expand_dims(self.initial_seasonals, axis=-1)

        init_params = {param: self.params[param] for param in initialize_params}

        params = self.__estimate_params(
                        init_components = [self.initial_level, self.initial_trend, self.initial_seasonals, self.seasonal_periods],
                        params = list(init_params.values()))
        
        for index, param in enumerate(params):

            init_params[list(init_params.keys())[index]] = param

        for param in self.params:

            if param in init_params:
                self.params[param] = init_params[param]

        if not self.damped:
            self.params["phi"] = 1

    def calculate_conf_level(self, conf):
        """Calculate the confidence level to be used for prediction intervals"""

        return round(stats.norm.ppf(1 - ((1 - conf) / 2)), 2)
    
    def update_res_variance(self, residuals, error):
        """Update the variance of the residuals during fitting"""
        residuals = torch.cat((residuals, torch.reshape(error, (1,1))))
        res_mean = torch.sum(residuals)/residuals.shape[0]
        residual_variance = torch.sum(torch.square(torch.sub(residuals, res_mean)))
        residual_variance = torch.divide(residual_variance, residuals.shape[0]-1)

        return residuals, res_mean, residual_variance
    
    def future_sample_paths(self, h, confidence):
        """
        Future path sampling for ETS models with no known equations for generating prediction intervals
        Errors are assumed to be normally distributed and random future paths are sampled from the normal distribution
        with mean and variance calculated by the residuals during fitting 
        """
        q1 = (1-confidence)/2
        q2 = 1 - q1

        loc = self.residual_mean
        scale = torch.sqrt(self.residual_variance)

        sample_paths = torch.tensor([])

        for iter in range(5000):

            sample = torch.normal(loc, scale, size=(1,h))
            sample_paths = torch.cat((sample_paths, sample))

        q1_sample = torch.quantile(sample_paths, q1, dim=0, interpolation="nearest")
        q2_sample = torch.quantile(sample_paths, q2, dim=0, interpolation="nearest")

        bounds = torch.abs(torch.sub(q1_sample, q2_sample))

        return bounds







import numpy as np
import pandas as pd
import torch
import scipy.optimize as opt
from scipy.stats import norm
from scipy.stats import linregress
from exponential_smoothing.initialization import ets_methods, smoothing_methods
from preprocessing.dataloader import DataLoader

def get_init_method(method):

    init_method = {"heuristic": heuristic_initialization,
                    "mle": mle_initialization}[method]
    
    return init_method

def get_smooth_method(error, trend, seasonal, damped):

    if error:
        smooth_method = { 
                            (None, False, None, "add"): ets_methods.ANN,
                            (None, False, "add", "add"): ets_methods.ANA,
                            (None, False, "mul", "add"): ets_methods.ANM,
                            ("add", False, None, "add"): ets_methods.AAN,
                            ("add", False, "add", "add"): ets_methods.AAA,
                            ("add", False, "mul", "add"): ets_methods.AAM,
                            ("add", True, None, "add"): ets_methods.AAdN,
                            ("add", True, "add", "add"): ets_methods.AAdA,
                            ("add", True, "mul", "add"): ets_methods.AAdM,

                            (None, False, None, "mul"): ets_methods.MNN,
                            (None, False, "add", "mul"): ets_methods.MNA,
                            (None, False, "mul", "mul"): ets_methods.MNM,
                            ("add", False, None, "mul"): ets_methods.MAN,
                            ("add", False, "add", "mul"): ets_methods.MAA,
                            ("add", False, "mul", "mul"): ets_methods.MAM,
                            ("add", True, None, "mul"): ets_methods.MAdN,
                            ("add", True, "add", "mul"): ets_methods.MAdA,
                            ("add", True, "mul", "mul"): ets_methods.MAdM
                        }[trend, damped, seasonal, error]
    
    else:
        smooth_method = { 
                        (None, False, None): smoothing_methods.simple_exp,
                        ("add", False, None): smoothing_methods.holt_trend,
                        ("add", True, None): smoothing_methods.holt_damped_trend,
                        ("add", False, "add"): smoothing_methods.hw_add,
                        ("add", False, "mul"): smoothing_methods.hw_mul,
                        ("add", True, "add"): smoothing_methods.hw_damped_add,
                        ("add", True, "mul"): smoothing_methods.hw_damped_mul
                        }[trend, damped, seasonal]
        
    return smooth_method

def heuristic_initialization(data, trend=False, seasonal=None,
                              seasonal_periods=None):
    """
    Heuristic initialization method for initial components
    See: Hyndman et al. section 2.6.1
    """

    data = DataLoader(data).to_numpy().copy()

    if data.ndim >= 1:

        data = np.squeeze(data)

    initial_level = None
    initial_trend = None
    initial_seasonal = None

    assert(len(data) >= 10), "Length of data must be >= for heuristic initialization"

    if seasonal:
        
        #Data must have at least 2 full seasonal cycles
        assert (len(data) > 2*seasonal_periods), "Length of data must be > 2*seasonal_periods"

        #Max number of seasonal cycles to be used is 5
        seasonal_cycles = min(5, len(data)//seasonal_periods)
        series = pd.Series(data[:seasonal_periods*seasonal_cycles])
        moving_avg = series.rolling(seasonal_periods, center=True).mean()

        if seasonal_periods % 2 == 0:
            moving_avg = moving_avg.shift(-1).rolling(2).mean()

        if seasonal == "add":
            detrend = series - moving_avg
        elif seasonal == "mul":
            detrend = series/moving_avg
     
        initial_seasonal = np.zeros(shape=(seasonal_periods,seasonal_cycles))*np.nan
        for i in range(0, seasonal_cycles):
            initial_seasonal[:, i] = detrend[i*seasonal_periods:(i+1)*seasonal_periods]
    
        initial_seasonal = np.nanmean(initial_seasonal, axis=1)

        if seasonal == "add":
            #normalize so that the sum is equal to 1
            initial_seasonal /= np.sum(initial_seasonal)
        elif seasonal == "mul":
            #normalize so that the sum is equal to m=seasonal_periods
            initial_seasonal *= seasonal_periods/np.sum(initial_seasonal)

        adjusted_data = moving_avg.dropna().values
    
    else:
        adjusted_data = data.copy()

    
    result = linregress(x=np.arange(10), y=adjusted_data[:10])
    initial_level = result[1]

    if trend:
        initial_trend = result[0]

    return initial_level, initial_trend, initial_seasonal


def mle_initialization(data, trend=None,damped=False, seasonal=None, error_type=None,seasonal_periods=12,alpha=0.1,beta=0.01,gamma=0.01,phi=0.99):
    """Maximum Likelihood Estimatin for initialization of initial components"""
 
    data = DataLoader(data).to_numpy().copy()
    selected_model = get_smooth_method(error=error_type, trend=trend, seasonal=seasonal, damped=damped)

    # Find the parameters of selected model to maximize the likelihood function.
    params = {
        (None, False, None): [alpha],
        (None, False, "add"): [alpha, gamma],
        (None, False, "mul"): [alpha, gamma],
        ("add", False, None): [alpha, beta],
        ("add", False, "add"): [alpha, beta, gamma],
        ("add", False, "mul"): [alpha, beta, gamma],
        ("add", True, None): [alpha, beta, phi],
        ("add", True, "add"): [alpha, beta, gamma, phi],
        ("add", True, "mul"): [alpha, beta, gamma, phi],
    }
    model_params = params[(trend, damped, seasonal)]
    seasonals = np.zeros(seasonal_periods+1).tolist() # Seasonal components are initialized as zero.
    
    # Flattening the components and parameters to use in optimization function.
    init_components = [data.mean(), 0] #level and trend
    if seasonal:
        for i in range(seasonal_periods):
            init_components.append(1) #initialize seasonals as 1
        
    init_components.append(seasonal_periods)
    init_values = init_components + model_params

    def log_likelihood(init_values,data,seasonal_periods,model):
        # Split the initial values to components, seasonal components and parameters.
        components = init_values[:2].tolist()
        seasonal_components = init_values[2:-len(model_params)].tolist()
        params = init_values[-len(model_params):]
        components = components + [seasonal_components] + [seasonal_periods]

        dep_var = DataLoader(data).to_tensor().detach().clone()
        if dep_var.ndim == 1:
            dep_var = torch.unsqueeze(dep_var, axis=1)
        # Calculate the log likelihood of the model.
        errors = model(dep_var,init_components=components,params=params)
        loc = np.array(errors).mean()
        scale = np.array(errors).std()**2
        log_likelihood = np.sum(norm.logpdf(errors, loc=loc, scale=scale))
        return -log_likelihood

    # Use the optimization function to find the maximum likelihood estimates with random initial values
    result_params = opt.minimize(log_likelihood,init_values, args=(data,seasonal_periods,selected_model,))

    def parse_results(result_params):
        # Unflatten the components and parameters.
        results = list(result_params.x)
        initial_level, initial_trend = results[:2]
        initial_seasonal = np.array(results[2:-len(model_params)-1])
        if not trend:
            initial_trend = None
        if not seasonal:
            initial_seasonal = None
        init_components = (initial_level, initial_trend, initial_seasonal)
        return init_components
    
    result_components = parse_results(result_params)

    return result_components


from .model import Smoothing_Model
from exponential_smoothing.models.smoothing import *


class ExponentialSmoothing:

    def __new__(self, dep_var, trend=None, damped=False, seasonal=None, seasonal_periods=None, initialization_method="heuristic", **kwargs):
        """
        Exponential Smoothing model for time series data

        Arguments:

        *dep_var (array_like): Univariate time series data
        *trend (Optional[str]): Trend component; None or "add"
        *damped (bool): Damp factor of the trend component; False if trend is None
        *seasonal (Optional[str]): Seasonal component; None, "add" or "mul"
        *seasonal_periods (Optional[int]): Cyclic period of the seasonal component; int or None if seasonal is None
        *initialization_method (str): Initialization method to use for the model parameters; "heuristic" or "mle"

        Keyword Arguments:

        ** alpha (float): Smoothing parameter for level component; takes values in (0,1)
        ** beta (float): Smoothing parameter for trend component; takes values in (0,1)
        ** phi (float): Damp factor for trend component; takes values in (0,1]
        ** gamma (float): Smoothing parameter for seasonal component; takes values in (0,1)
        
        ETS models are implemented by the below textbook as a reference:
        'Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
        and practice. OTexts, 2014.'
        """
    
        smoothing_class = { 
                        (None, None): SES,
                        ("add", None): HoltTrend,
                        ("add", "add"): HoltWinters,
                        ("add", "mul"): HoltWinters,
                                                    }[trend, seasonal]

        return smoothing_class(dep_var, trend=trend, seasonal=seasonal, damped=damped, 
                               seasonal_periods=seasonal_periods, initialization_method=initialization_method, **kwargs)



import matplotlib
import matplotlib.style
import matplotlib.pyplot as plt
from preprocessing.dataloader import DataLoader
from utils.evaluation_utils import * 

def plot_predictions(y_true, y_pred, bounds=None, pre_vals=None, figsize=(12,8), colors=None, bounds_fill_alpha=0.7, title=None, style=None, metrics=None):
    """
    Utility function for plotting prediction results of the time series model
    
    Arguments:

    *y_true (array_like): Ground truth values
    *y_pred (array_like): Predicted values
    *bounds (Optional[iterable]): Confidence bounds for the prediction interval
    *pre_vals (Optional[array_like]): Values that come before the predicted values i.e; last n-points of the training data
    *figsize (Optional[tuple]): Size of the plot
    *colors (Optional[iterable]): Colors of the lines/points on the plot
    *bounds_fill_alpha (Optional[float]): Alpha for the transparency of the filled values between confidence bounds
    *title (Optional[str]): Title of the plot
    *style (Optional[str]): Style of the plot 'https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html'
    *metrics (Optional[iterable]): Evaluation metrics to use for the predictions to report on the plot
    """
    y_true = DataLoader(y_true).to_numpy()
    y_pred = DataLoader(y_pred).to_numpy()

    use_colors = {"y_true": "blue", "y_pred": "orange", "bounds": "gray", "pre_vals": "black"}

    if colors:
        try:
            iter(colors)
        except TypeError:
            raise TypeError("Provide colors as an iterable")
        
        if type(colors) == dict:
            for key in colors:
                assert(key in list(use_colors.keys())), f"Ensure that keys in colours dictionary are {list(use_colors.keys())}"
                use_colors[key] = colors[key]

        else:
            for ind, c in enumerate(colors):
                use_colors[list(use_colors.keys())[ind]] = c

    error_bounds = None

    if bounds:
        try:
            iter(bounds)
        except TypeError:
            raise TypeError("Provide bounds as an iterable of length 2")
        
        assert (len(bounds) == 2), "Provide bounds as an iterable of length 2"

        if type(bounds) == dict:

            assert (list(bounds.keys()) == ["upper", "lower"]), "Provide bounds dictionary keys as ['upper', 'lower']"

            for val in list(bounds.values()):
                assert (len(val) == len(y_pred)), "Length of bounds must match length of predictions"

            error_bounds = {"upper": DataLoader(bounds["upper"]).to_numpy(), "lower": DataLoader(bounds["lower"]).to_numpy()}

        else:
            for val in bounds:
                assert (len(val) == len(y_pred)), "Length of bounds must match length of predictions"

            error_bounds = {"upper": DataLoader(bounds[0]).to_numpy(), "lower": DataLoader(bounds[1]).to_numpy()}
    
    plt_metrics = None
    if metrics:
        plt_metrics = {}
        try:
            iter(metrics)
            if type(metrics) == dict:
                raise TypeError("Metrics argument cannot be a dictionary")
        except TypeError:
            raise TypeError("Provide metrics as an iterable of length 2")
        
        for i in metrics:
            assert (type(i) == str), "Provide metrics as an iterable with string entries"
            assert (i in ["rmse", "mse", "mae"]), "Supported metrics are: ['rmse', 'mse' and 'mae']"

            if i == "rmse":
                plt_metrics["RMSE"] = rmse(y_pred, y_true).item()
            if i == "mse":
                plt_metrics["MSE"] = mse(y_pred, y_true).item()
            if i == "mae":
                plt_metrics["MAE"] = mae(y_pred, y_true).item()
    
    assert (type(title) == str or title is None), "Plot title must be a string"

    if style:
        assert(type(style) == str), "Provide style as a string"
        matplotlib.style.use(style)

    plt.figure(figsize=figsize)
    if pre_vals is not None:
        pre_vals = DataLoader(pre_vals).to_numpy()
        main_plt_range = range(len(pre_vals), len(pre_vals)+len(y_pred))
        plt.plot(range(len(pre_vals)), pre_vals, color=use_colors["pre_vals"])
    else:
        main_plt_range = range(len(y_pred))
    plt.plot(main_plt_range, y_true, color=use_colors["y_true"], label="Y True")
    plt.plot(main_plt_range, y_pred, color=use_colors["y_pred"], label="Y Predicted")
    if error_bounds:
        plt.fill_between(main_plt_range, error_bounds["upper"], error_bounds["lower"], 
                         color=use_colors["bounds"], alpha=bounds_fill_alpha, label="Prediction Error Bounds")
    if plt_metrics:
        append_title = ""
        for metric in plt_metrics:
            score = plt_metrics[metric]
            append_title += metric + f":{score:.3f} "
        if title:
            title = title + "\n" + append_title
        else:
            title = append_title
    plt.title(title)
    plt.legend(loc="best")
    plt.show()
    


from .model import ETS_Model
from exponential_smoothing.models.ets_models import *

class ETS(ETS_Model):

    def __new__(self, dep_var, error_type="add", trend=None, damped=False, seasonal=None, seasonal_periods=None, initialization_method="heuristic", **kwargs):
        """
        ETS (Error,Trend,Seasonality) model for time series data

        Arguments:

        *dep_var (array_like): Univariate time series data
        *error_type (str): Type of error of the ETS model; "add" or "mul"
        *trend (Optional[str]): Trend component; None or "add"
        *damped (bool): Damp factor of the trend component; False if trend is None
        *seasonal (Optional[str]): Seasonal component; None, "add" or "mul"
        *seasonal_periods (Optional[int]): Cyclic period of the seasonal component; int or None if seasonal is None
        *initialization_method (str): Initialization method to use for the model parameters; "heuristic" or "mle"

        Keyword Arguments:

        ** alpha (float): Smoothing parameter for level component; takes values in (0,1)
        ** beta (float): Smoothing parameter for trend component; takes values in (0,1)
        ** phi (float): Damp factor for trend component; takes values in (0,1]
        ** gamma (float): Smoothing parameter for seasonal component; takes values in (0,1)
        
        ETS models are implemented by the below textbook as a reference:
        'Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
        and practice. OTexts, 2014.'
        """
        
        ets_class = { 
                        (None, None, "add"): ETS_ANN,
                        (None, "add", "add"): ETS_ANA,
                        (None, "mul", "add"): ETS_ANM,
                        ("add",  None, "add"): ETS_AAN,
                        ("add", "add", "add"): ETS_AAA,
                        ("add", "mul", "add"): ETS_AAM,

                        (None,  None, "mul"): ETS_MNN,
                        (None,  "add", "mul"): ETS_MNA,
                        (None, "mul", "mul"): ETS_MNM,
                        ("add",  None, "mul"): ETS_MAN,
                        ("add", "add", "mul"): ETS_MAA,
                        ("add", "mul", "mul"): ETS_MAM,
                                                            }[trend, seasonal, error_type]

        return ets_class(dep_var, trend=trend, seasonal=seasonal, error_type=error_type, damped=damped, 
                         seasonal_periods=seasonal_periods, initialization_method=initialization_method, **kwargs)



import numpy as np

"""ETS methods to use for initialization of parameters"""

def simple_exp(dep_var, init_components, params):

    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components          

    alpha = params[0]

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            errors.append(0)

        else:               

            y_hat = lvl
            error = row.numpy()[0] - y_hat
            

            lprev = lvl
            lvl = alpha*row.numpy()[0] + (1-alpha)*lprev

            errors.append(error)

    return errors

def holt_trend(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            errors.append(0) 

        else:               

            y_hat = lvl + trend 
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            lvl = alpha*row.numpy()[0] + (1-alpha)*(lprev+bprev)
            trend = beta*(lvl-lprev) + (1-beta)*bprev

            errors.append(error)

    return errors

def holt_damped_trend(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            errors.append(0) 

        else:               

            y_hat = lvl + phi*trend 
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            lvl = alpha*row.numpy()[0] + (1-alpha)*(lprev+phi*bprev)
            trend = beta*(lvl-lprev) + (1-beta)*phi*bprev

            errors.append(error)

    return errors

def hw_add(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + trend + seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]-lprev-bprev)
            lvl = alpha*(row.numpy()[0]-seasonal) + (1-alpha)*(lprev+ bprev)
            trend = beta*(lvl - lprev) + (1-beta)*bprev

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + trend + seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]-lprev-bprev)
            lvl = alpha*(row.numpy()[0]-seasonal) + (1-alpha)*(lprev+ bprev)
            trend = beta*(lvl - lprev) + (1-beta)*bprev

            seasonals.append(seasonal_t)
            errors.append(error)         

    return errors

def hw_damped_add(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = lvl + phi*trend + seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]-lprev-phi*bprev)
            lvl = alpha*(row.numpy()[0]-seasonal) + (1-alpha)*(lprev + phi*bprev)
            trend = beta*(lvl - lprev) + (1-beta)*phi*bprev

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = lvl + phi*trend + seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]-lprev-phi*bprev)
            lvl = alpha*(row.numpy()[0]-seasonal) + (1-alpha)*(lprev + phi*bprev)
            trend = beta*(lvl - lprev) + (1-beta)*phi*bprev

            seasonals.append(seasonal_t)
            errors.append(error)             

    return errors

def hw_mul(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = (lvl + trend)*seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]/(lprev+bprev))
            lvl = alpha*(row.numpy()[0]/seasonal) + (1-alpha)*(lprev + bprev)
            trend = beta*(lvl - lprev) + (1-beta)*bprev

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = (lvl + trend)*seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]/(lprev+bprev))
            lvl = alpha*(row.numpy()[0]/seasonal) + (1-alpha)*(lprev + bprev)
            trend = beta*(lvl - lprev) + (1-beta)*bprev

            seasonals.append(seasonal_t)
            errors.append(error)            
        

    return errors

def hw_damped_mul(dep_var, init_components, params):
    
    init_lvl, init_trend, init_seasonals, seasonal_periods = init_components           
    alpha, beta, gamma, phi = params

    errors = []

    for index, row in enumerate(dep_var):

        if index == 0:

            lvl = init_lvl
            trend = init_trend
            seasonal = init_seasonals[0]
            seasonals = [seasonal]
            errors.append(0)

        elif index < seasonal_periods:            

            seasonal = init_seasonals[index]     

            y_hat = (lvl + phi*trend)*seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]/(lprev+phi*bprev))
            lvl = alpha*(row.numpy()[0]/seasonal) + (1-alpha)*(lprev + phi*bprev)
            trend = beta*(lvl - lprev) + (1-beta)*phi*bprev

            seasonals.append(seasonal_t)
            errors.append(error)            

        else:               

            seasonal = seasonals[index-seasonal_periods]  

            y_hat = (lvl + phi*trend)*seasonal
            error = row.numpy()[0] - y_hat

            lprev, bprev = lvl, trend

            seasonal_t = (1-gamma)*seasonal + gamma*(row.numpy()[0]/(lprev+phi*bprev))
            lvl = alpha*(row.numpy()[0]/seasonal) + (1-alpha)*(lprev + phi*bprev)
            trend = beta*(lvl - lprev) + (1-beta)*phi*bprev

            seasonals.append(seasonal_t)
            errors.append(error)                     

    return errors


from preprocessing.dataloader import DataLoader
from utils.vis_utils import plot_decomp
import numpy as np
import pandas as pd

def classical_decomposition(data, seasonal_period, method = 'add', show = True):
    """
    Classical Decomposition for univariate time series data

    Arguments:

    *data (array_like): Time series data to perform decomposition on
    *seasonal_period (int): Seasonal period of the given data
    *method (Optional[str]): Decomposition method to be used; "add" or "mul"
    *show (Optional[bool]): Whether to plot the decomposition results

    Chapter 6.3 of the textbook is taken as a reference:
    'Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
    and practice. OTexts, 2014.'
    """
    data = DataLoader(data).to_numpy().copy()

    if data.ndim >= 1:

        data = np.squeeze(data)

    number_of_cycles = len(data) // seasonal_period
    assert (number_of_cycles >= 2), "Data must have at least 2 full seasonal cycles"

    # Computing the trend-cycle component using moving averages
    trend = pd.Series(data).rolling(seasonal_period,center=True).mean()

    if seasonal_period % 2 == 0:
        trend = trend.shift(-1).rolling(2).mean()
    
    # Detrending
    if method == 'add':
        detrended = data - trend.values
    elif method == 'mul':
        detrended = data / trend.values

    # Calculating the seasonal component
    seasonal = np.zeros(shape=(seasonal_period, number_of_cycles))*np.nan
    for i in range(0, number_of_cycles):
            seasonal[:, i] = detrended[i*seasonal_period:(i+1)*seasonal_period]
    seasonal = np.nanmean(seasonal, axis=1)
    
    # Putting seasonality component in a numpy array same length as the original data
    seasonal = np.array([seasonal[i%seasonal_period] for i in range(len(data))])
    
    # Deseasoning (Calculating remainder)
    if method == 'add' :
        remainder = detrended - seasonal
    elif method == 'mul':
        remainder = detrended / seasonal

    if show:
        plot_decomp(trend, seasonal, remainder)
    
    return trend, seasonal, remainder




import numpy as np
import torch
from exponential_smoothing.model import Smoothing_Model

"""
Exponential Smoothing models for time series forecasting.
All methods have been implemented from chapter 7 of the textbook as a reference.
'Hyndman, Rob J., and George Athanasopoulos. Forecasting: principles
and practice. OTexts, 2014.'
"""

class SES(Smoothing_Model):
    def __init__(self, dep_var, trend=None, seasonal=None, seasonal_periods=None, damped=False, indep_var=None, **kwargs):
        """Implementation of Simple Exponential Smoothing"""
        self.trend = None
        self.damped = False
        self.seasonal = None
        self.seasonal_periods = None

        super().set_allowed_kwargs(["alpha", "beta", "phi", "gamma"])
        
        super().__init__(dep_var, self.trend, self.seasonal, self.seasonal_periods, self.damped, **kwargs)
        
        initialize_params = ["alpha"]
        self.__initialize_params(initialize_params)

    def __initialize_params(self, initialize_params):
        """Initialize the model parameters"""
        super().initialize_params(initialize_params)

        self.initial_level = torch.tensor(self.init_components["level"])

        self.alpha, self.beta, self.gamma, self.phi = torch.tensor(list(self.params.values()), dtype=torch.float32)

    def __smooth_level(self, y, lprev):
        """Calculate level"""
        self.level =  torch.add(torch.mul(self.alpha, y),torch.mul((1 - self.alpha), lprev))
    
    def fit(self):
        """
        Fit the model to the data according to the equations:
        l_t =  alpha*y_t + (1 - alpha)*l_{t-1}
        """
        self.fitted = torch.zeros(self.dep_var.shape[0])
        for index,row in enumerate(self.dep_var):
            if index == 0:
                self.level = self.initial_level
                self.fitted[0] = row
            else:
                y_hat = self.level
                lprev = self.level
                self.__smooth_level(row, lprev)
                self.fitted[index] = y_hat  
    
    def predict(self,h):
        """
        Predict the next h values of the target variable.

        Arguments:

        *h (int): Future time-steps to forecast
        """
        self.forecast = torch.tensor([])
        for i in range(1,h+1):
            step_forecast = self.level
            self.forecast = torch.cat((self.forecast,step_forecast))
        return self.forecast


class HoltTrend(Smoothing_Model):
    def __init__(self, dep_var, trend="add", seasonal=None, seasonal_periods=None, damped=False, indep_var=None, **kwargs):
        """Implementation of Holt's Trend Method"""
        self.trend = "add"
        self.damped = damped
        self.seasonal = None
        self.seasonal_periods = None

        super().set_allowed_kwargs(["alpha", "beta", "phi", "gamma"])
        super().__init__(dep_var, self.trend, self.seasonal, self.seasonal_periods, self.damped, **kwargs)
        
        initialize_params = ["alpha", "beta"]
        if self.damped:
            initialize_params.append("phi")

        self.__initialize_params(initialize_params)


    def __initialize_params(self, initialize_params):
        """Initialize the model parameters"""

        super().initialize_params(initialize_params)

        self.initial_level = torch.tensor(self.init_components["level"])
        self.initial_trend = torch.tensor(self.init_components["trend"])

        self.alpha, self.beta, self.gamma, self.phi = torch.tensor(list(self.params.values()), dtype=torch.float32)

    def __smooth_level(self, y, lprev, bprev):
        """Calculate level"""
        self.level = torch.mul(torch.sub(1,self.alpha),torch.add(lprev, torch.mul(self.phi, bprev)))
        self.level = torch.add(torch.mul(self.alpha, y), self.level)

    def __smooth_trend(self, lprev, bprev):
        "Calculate trend"
        self.trend = torch.mul(self.beta, torch.sub(self.level, lprev))
        self.trend = torch.add(self.trend, torch.mul(torch.sub(1, self.beta), torch.mul(self.phi, bprev)))

    def fit(self):
        """
        Fit the model to the data according to the equations:
        l_t = alpha * y_t + (1 - alpha) * (l_{t-1} + b_{t-1})
        b_t = beta * (l_t - l_{t-1}) + (1 - beta)*phi*b_{t-1}
        """
        
        self.fitted = torch.zeros(size=self.dep_var.shape)
        for index, row in enumerate(self.dep_var):
            if index == 0:
                self.level = self.initial_level
                self.trend = self.initial_trend
                self.fitted[0] = row[0]
            else:
                y_hat = torch.add(self.level, torch.mul(self.phi, self.trend))
                lprev, bprev = self.level, self.trend
                self.__smooth_level(row, lprev, bprev)
                self.__smooth_trend(lprev, bprev)
                self.fitted[index] = y_hat
        

    def predict(self, h):
        """
        Predict the next h values of the target variable.

        Arguments:

        *h (int): Future time-steps to forecast
        """
        
        self.forecast = torch.tensor([])
        for i in range(1,h+1):

            damp_factor = torch.sum(torch.tensor([torch.pow(self.phi, x+1) for x in range(i)]))

            step_forecast = torch.add(self.level, torch.mul(damp_factor, self.trend))
            self.forecast = torch.cat((self.forecast, step_forecast))
            
        return self.forecast
    
class HoltWinters(Smoothing_Model):
    def __init__(self, dep_var, trend="add", seasonal="add", seasonal_periods=4, damped=False, indep_var=None, **kwargs):
        "Implementation of Holt-Winters' Seasonal Method"
        self.trend = trend
        self.damped = damped
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods

        super().set_allowed_kwargs(["alpha", "beta", "phi", "gamma"])
        super().__init__(dep_var, self.trend, self.seasonal, self.seasonal_periods, self.damped, **kwargs)
        
        initialize_params = ["alpha", "beta", "gamma"]
        if self.damped:
            initialize_params.append("phi")

        self.__initialize_params(initialize_params)

    def __initialize_params(self, initialize_params):
        """Initialize the model parameters"""

        super().initialize_params(initialize_params)

        self.initial_level = torch.tensor(self.init_components["level"])
        self.initial_trend = torch.tensor(self.init_components["trend"])
        self.initial_seasonals = torch.tensor(self.init_components["seasonal"])

        self.alpha, self.beta, self.gamma, self.phi = torch.tensor(list(self.params.values()), dtype=torch.float32)


    def __smooth_level(self, y, lprev, bprev, seasonal):
        """Calculate level"""

        if self.seasonal == "add":

            self.level = torch.mul(torch.sub(1,self.alpha),torch.add(lprev, torch.mul(self.phi, bprev)))
            self.level = torch.add(torch.mul(self.alpha, torch.sub(y, seasonal)), self.level)
        
        elif self.seasonal == "mul":

            self.level = torch.mul(torch.sub(1,self.alpha),torch.add(lprev, torch.mul(self.phi, bprev)))
            self.level = torch.add(torch.mul(self.alpha, torch.divide(y, seasonal)), self.level)


    def __smooth_trend(self, lprev, bprev):
        """Calculate trend"""

        self.trend = torch.mul(self.beta, torch.sub(self.level, lprev))
        self.trend = torch.add(self.trend, torch.mul(torch.sub(1, self.beta), torch.mul(self.phi, bprev)))

    def __smooth_seasonal(self,  y, lprev, bprev, seasonal):
        '''This function calculates the smoothed trend for a given beta and level values'''
        
        if self.seasonal == "add":

            seasonal = torch.mul(torch.sub(1,self.gamma), seasonal)
            seasonal = torch.add(seasonal, torch.mul(self.gamma, torch.sub(torch.sub(y, lprev), torch.mul(self.phi, bprev))))
            self.seasonals = torch.cat((self.seasonals, seasonal))

        elif self.seasonal == "mul":

            seasonal = torch.mul(torch.sub(1,self.gamma), seasonal)
            seasonal = torch.add(seasonal, torch.mul(self.gamma, torch.divide(y, torch.add(lprev,torch.mul(self.phi, bprev)))))
            self.seasonals = torch.cat((self.seasonals, seasonal))


    def fit(self):
        """
        Fit the model to the data according to the equations:
        
        If seasonal is additive:
        l_t = alpha*(y_t - s_{t-m}) + (1 - alpha)*(l_{t-1} + b_{t-1})
        b_t = beta * (l_t - l_{t-1}) + (1 - beta)*phi*b_{t-1}
        s_t = gamma*(y_t - l_{t-1} - phi*b_{t-1}) + (1 - y)*s_{t-m}

        If seasonal is multiplicative:
        l_t = alpha*(y_t/s_{t-m}) + (1 - alpha)*(l_{t-1} + b_{t-1})
        b_t = beta * (l_t - l_{t-1}) + (1 - beta)*phi*b_{t-1}
        s_t = gamma*(y_t/(l_{t-1} + phi*b_{t-1})) + (1 - y)*s_{t-m}
        """
        
        self.fitted = torch.zeros(size=self.dep_var.shape)

        for index, row in enumerate(self.dep_var):

            if index == 0:
                self.level = self.initial_level
                self.trend = self.initial_trend
                seasonal = self.initial_seasonals[0]
                self.seasonals = torch.tensor([seasonal])
                self.fitted[0] = row
            
            elif index < self.seasonal_periods:

                seasonal = self.initial_seasonals[index]
                
                if self.seasonal == "add":
                    y_hat = torch.add(seasonal, torch.add(self.level, torch.mul(self.phi, self.trend)))
                elif self.seasonal == "mul":
                    y_hat = torch.mul(seasonal, torch.add(self.level, torch.mul(self.phi, self.trend)))

                lprev, bprev = self.level, self.trend
                self.__smooth_level(row, lprev, bprev, seasonal)
                self.__smooth_trend(lprev, bprev)

                self.seasonals = torch.cat((self.seasonals, seasonal))

                self.fitted[index] = y_hat
                
            
            else:

                seasonal = self.seasonals[-self.seasonal_periods]
                
                if self.seasonal == "add":
                    y_hat = torch.add(seasonal, torch.add(self.level, torch.mul(self.phi, self.trend)))
                elif self.seasonal == "mul":
                    y_hat = torch.mul(seasonal, torch.add(self.level, torch.mul(self.phi, self.trend)))

                lprev, bprev = self.level, self.trend
                self.__smooth_level(row, lprev, bprev, seasonal)
                self.__smooth_trend(lprev, bprev)
                self.__smooth_seasonal(row, lprev, bprev, seasonal)

                self.fitted[index] = y_hat
            

    def predict(self, h):
        """
        Predict the next h values of the target variable.

        Arguments:

        *h (int): Future time-steps to forecast
        """
        
        self.forecast = torch.tensor([])
        len_seasonal = self.seasonals.shape[0]
        for i in range(1,h+1):
       
            damp_factor = torch.sum(torch.tensor([torch.pow(self.phi, x+1) for x in range(i)]))
            k = int((i-1)/self.seasonal_periods)
   
            step_forecast = torch.add(self.level, torch.mul(damp_factor, self.trend))

            if self.seasonal == "mul":
                step_forecast = torch.mul(step_forecast, self.seasonals[len_seasonal+i-self.seasonal_periods*(k+1)])
            elif self.seasonal == "add":
                step_forecast = torch.add(step_forecast, self.seasonals[len_seasonal+i-self.seasonal_periods*(k+1)])
            
            self.forecast = torch.cat((self.forecast, step_forecast))
            
        return self.forecast



            
